package winterwell.utils;

import java.math.BigInteger;
import java.util.Collection;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import winterwell.utils.containers.Containers;

/**
 * @testedby {@link MathUtilsTest}
 * @author daniel
 * 
 */
public class MathUtils {

	private static double MACH_EPS = -1;

	private static Pattern NUMBER = Pattern
			.compile("[^0-9\\-]?([0-9\\.\\-]+)[^0-9]{0,3}");

	public static final double ROOT_TWO = Math.sqrt(2);

	public static double[] abs(double... values) {
		double[] values2 = new double[values.length];
		for (int i = 0; i < values.length; i++) {
			values2[i] = Math.abs(values[i]);
		}
		return values2;
	}

	/**
	 * A less precise version of {@link #equalish(double, double)}. Useful for
	 * testing stochastic things.
	 * 
	 * @param x
	 * @param y
	 * @return true if x roughly equals y. Uses ~25% tolerance for non-zero x,y.
	 *         The result is inherently arbitrary if x or y are zero.
	 */
	public static boolean approx(double x, double y) {
		if (x == 0 || y == 0)
			// arbitrary limit
			return Math.abs(x - y) < 0.25;
		double d = Math.abs(x - y);
		double m2 = Math.abs(x + y);
		return d / m2 < 0.125;
	}

	/**
	 * @param x
	 * @param y
	 * @return true if x approximately equals y. Uses ~1% tolerance for non-zero
	 *         x,y. The result is inherently arbitrary if x or y are zero.
	 */
	public static boolean equalish(double x, double y) {
		if (x == 0 || y == 0)
			// arbitrary limit
			return Math.abs(x - y) < 0.001;
		double d = Math.abs(x - y);
		double m2 = Math.abs(x + y);
		return d / m2 < 0.005;
	}

	public static double euclideanLength(double[] vector) {
		double x2 = 0;
		for (double d : vector) {
			x2 += d * d;
		}
		return Math.sqrt(x2);
	}

	/**
	 * Simple convenience function: Cap value - if necessary - to lie in [min,
	 * max].
	 * 
	 * @param x
	 * @param min
	 * @param max
	 * @return x'
	 */
	public static double forceIn(double x, double min, double max) {
		assert min <= max;
		if (x < min)
			return min;
		if (x > max)
			return max;
		return x;
	}

	/**
	 * Machine epsilon gives an upper bound on the relative error due to
	 * rounding in floating point arithmetic.
	 * 
	 * @author Wikipedia
	 */
	public static double getMachineEpsilon() {
		if (MACH_EPS != -1)
			return MACH_EPS;
		double machEps = 1.0;
		do {
			machEps /= 2.0;
		} while ((1.0 + (machEps / 2.0)) != 1.0);
		MACH_EPS = machEps;
		return MACH_EPS;
	}

	/**
	 * @param num
	 *            Can be null. This can include an optional leading symbol and
	 *            upto 3 trailing symbols (any non digit character or point).
	 *            E.g. £10 (but not -£10, at least for now), 20EUR, 100px, 0.5%,
	 *            etc.
	 * @return double. 0 if num could not be parsed. Does not throw an exception
	 *         if num is null or badly formatted.
	 * @see StrUtils#isNumber(String)
	 * @testedby {@link MathUtilsTest#testGetNumber()}
	 */
	public static double getNumber(String num) {
		if (num == null)
			return 0;
		Matcher m = NUMBER.matcher(num);
		boolean ok = m.matches();
		if (!ok)
			return 0;
		try {
			if (m.start(1) != 0 || m.end(1) != num.length()) {
				num = m.group(1);
			}
			return Double.valueOf(num);
		} catch (NumberFormatException e) {
			return 0;
		}
	}

	/**
	 * @param x
	 * @return true if not infinite and not NaN
	 * @testedby {@link MathUtilsTest#testIsFinite()}
	 */
	public static boolean isFinite(double x) {
		return !(Double.isNaN(x) || Double.isInfinite(x));
	}

	public static boolean isProb(double x) {
		return x >= 0 && x <= 1;
	}

	/**
	 * @param x
	 * @return if true, this should be treated as zero. Never divide by such a
	 *         number!
	 */
	public static boolean isTooSmall(double x) {
		return x < getMachineEpsilon();
	}

	/**
	 * @param a
	 * @param b
	 * @return true if a < b
	 */
	public static boolean lessThan(Number a, Number b) {
		// unavoidably ugly
		if (a instanceof BigInteger) {
			BigInteger bigB = b instanceof BigInteger ? (BigInteger) b
					: new BigInteger(b.toString());
			int c = ((BigInteger) a).compareTo(bigB);
			return c < 0;
		}
		if (b instanceof BigInteger) {
			BigInteger bigA = a instanceof BigInteger ? (BigInteger) a
					: new BigInteger(a.toString());
			int c = bigA.compareTo((BigInteger) b);
			return c < 0;
		}
		if (a instanceof Long || b instanceof Long)
			return a.longValue() < b.longValue();
		return a.doubleValue() < b.doubleValue();
	}

	/**
	 * 
	 * @param xs
	 *            Can contain nulls -- which are ignored
	 * @return this will be a member of the xs, null if the xs had no non-null
	 *         values
	 */
	public static <N extends Number> N max(Collection<N> xs) {
		assert !xs.isEmpty();
		double max = Double.NEGATIVE_INFINITY;
		N maxn = null;
		for (N number : xs) {
			// skip null!
			if (number == null) {
				continue;
			}
			double x = number.doubleValue();
			if (x > max) {
				max = x;
				maxn = number;
			}
		}
		return maxn;
	}

	/**
	 * 
	 * @param values
	 *            Must not be zero-length or null
	 * @return max of values
	 */
	public static double max(double... values) {
		double max = values[0];
		for (double i : values) {
			if (i > max) {
				max = i;
			}
		}
		return max;
	}

	public static int max(int... values) {
		int max = values[0];
		for (int i : values) {
			if (i > max) {
				max = i;
			}
		}
		return max;
	}

	/**
	 * 
	 * @param xs
	 *            Can contain nulls -- which are ignored
	 * @return this will be a member of the xs
	 */
	public static <N extends Number> N min(Collection<N> xs) {
		assert !xs.isEmpty();
		double min = Double.POSITIVE_INFINITY;
		N minn = null;
		for (N number : xs) {
			// skip null!
			if (number == null) {
				continue;
			}
			double x = number.doubleValue();
			if (x < min) {
				min = x;
				minn = number;
			}
		}
		return minn;
	}

	/**
	 * @param values
	 *            Must not be zero-length or null
	 * @return min of values (remember that -100 beats 1 - use min + abs if you
	 *         want the smallest number)
	 */
	public static double min(double... values) {
		double min = values[0];
		for (double i : values) {
			if (i < min) {
				min = i;
			}
		}
		return min;
	}

	/**
	 * @param x
	 * @return x*x (yes it's just a minor convenience)
	 */
	public static double sq(double x) {
		return x * x;
	}

	public static double sum(double[] values) {
		double total = 0;
		for (double d : values) {
			total += d;
		}
		return total;
	}

	/**
	 * Copy numbers into an array. ?? Should this be in {@link Containers}?
	 */
	public static double[] toArray(Collection<? extends Number> values) {
		double[] arr = new double[values.size()];
		int i = 0;
		for (Number d : values) {
			arr[i] = d.doubleValue();
			i++;
		}
		return arr;
	}

	/**
	 * Copy numbers into an array. Note that using a Trove list in the first
	 * place is more efficient. ?? Should this be in {@link Containers}?
	 */
	public static int[] toIntArray(Collection<Integer> values) {
		int[] arr = new int[values.size()];
		int i = 0;
		for (Integer d : values) {
			arr[i] = d;
			i++;
		}
		return arr;
	}

}
