package winterwell.utils;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import winterwell.utils.io.FileUtils;

/**
 * Convenience wrapper for the {@link java.lang.Process} class. Makes life more
 * pleasant and less buggy.
 * <p>
 * <i>Warning:</i> A common error is to think Java Process acts like a shell
 * interpreter. It does not. It runs a single executable (a program or script).
 * Things like pipes, redirects, and chained commands will not work. Try
 * {@link ShellScript} for experimental support of shell-script style commands.
 * </p>
 * See http://www.javaworld.com/javaworld/jw-12-2000/jw-1229-traps.html
 * 
 * TODO implement Closeable, and replace destroy()
 * 
 * @author daniel
 * 
 */
public class Process {

	boolean closed;

	private String command;

	private boolean echo;

	private StreamGobbler err;

	private StreamGobbler out;
	private final ProcessBuilder pb;

	private java.lang.Process process;

	public Process(List<String> command) {
		pb = new ProcessBuilder(command);
	}

	public Process(String command) {
		this.command = command;
		pb = new ProcessBuilder(parse(command));
	}

	/**
	 * Close the process' 3 pipes in, out and error. See
	 * http://www.vnoel.com/Computers
	 * -Related-Issues/Java-problem-Too-many-open-files.html Experimentation
	 * suggests this may not be a problem after all see
	 * {@link ProcessTest#testOpenFilesIssue()}
	 */
	private void closeStreams() {
		if (process == null)
			return;
		if (err != null) {
			err.pleaseStop();
		}
		if (out != null) {
			out.pleaseStop();
		}
		FileUtils.close(process.getInputStream());
		FileUtils.close(process.getOutputStream());
		FileUtils.close(process.getErrorStream());
	}

	/**
	 * Kills the subprocess. The subprocess represented by this
	 * <code>Process</code> object is forcibly terminated.
	 */
	public void destroy() {
		if (closed)
			return;
		closed = true;
		process.destroy();
		closeStreams();
	}

	/**
	 * Ensure that process streams are closed. Does not terminate the process.
	 * 
	 * @see java.lang.Object#finalize()
	 */
	@Override
	protected void finalize() throws Throwable {
		super.finalize();
		closeStreams();
	}

	public String getCommand() {
		return command == null ? Printer.toString(pb.command(), " ") : command;
	}

	/**
	 * Returns a string map view of this process builder's environment. The
	 * returned object may be modified using ordinary {@link java.util.Map Map}
	 * operations.
	 * 
	 * @see ProcessBuilder#environment()
	 */
	public Map<String, String> getEnvironment() {
		return pb.environment();
	}

	/**
	 * @return the process's error output, or "" if the redirectErrorStream has
	 *         been set to true.
	 */
	public String getError() {
		return err == null ? "" : err.getString();
	}

	public String getOutput() {
		return out.getString();
	}

	/**
	 * @deprecated Depends on poking around inside the Process object's private
	 *             parts.
	 * @return The UNIX process id (pid)
	 */
	@Deprecated
	public int getProcessId() {
		Integer id = ReflectionUtils.getPrivateField(process, "pid");
		return id;
	}

	/**
	 * Split a command up into the base call and respective arguments. Handles
	 * OS-specific hacks and other unpleasantries.
	 * 
	 * @param command
	 * @return E.g. "ls -l foo*" would be ["ls", "-l", "foo*"]
	 */
	List<String> parse(String command) {
		// support "quoted args" which should not be split up
		List<String> bits = new ArrayList<String>();
		StringBuilder bit = new StringBuilder();
		boolean inQuotes = false;
		for (char c : command.toCharArray()) {
			if (inQuotes) {
				bit.append(c);
				if (c == '"') {
					inQuotes = false;
				}
				continue;
			}
			if (Character.isWhitespace(c)) {
				if (bit.length() != 0) {
					bits.add(bit.toString());
					bit = new StringBuilder();
				}
				continue;
			}
			bit.append(c);
			if (c == '"') {
				inQuotes = true;
			}
		}
		if (bit.length() != 0) {
			bits.add(bit.toString());
		}
		// Windows requires we call cmd.exe
		String osName = System.getProperty("os.name");
		if (osName.equals("Windows 95")) {
			ArrayList<String> wbits = new ArrayList<String>(bits.size() + 2);
			wbits.add("command.com");
			wbits.add("/C");
			wbits.addAll(bits);
			bits = wbits;
		} else if (osName.startsWith("Windows")) {
			ArrayList<String> wbits = new ArrayList<String>(bits.size() + 2);
			wbits.add("cmd.exe");
			wbits.add("/C");
			wbits.addAll(bits);
			bits = wbits;
		}
		// Done
		return bits;
	}

	/**
	 * Execute the process. Returns immediately - use {@link #waitFor(long)} to
	 * block.
	 * 
	 * @throws IOException
	 */
	public void run() {
		try {
			process = pb.start();

			// Gobble output or Java hangs!
			out = new StreamGobbler(process.getInputStream());
			if (echo) {
				out.setEcho(true);
			}
			out.start();
			if (!pb.redirectErrorStream()) {
				err = new StreamGobbler(process.getErrorStream());
				if (echo) {
					err.setEcho(true);
				}
				err.start();
			}
		} catch (IOException e) {
			throw Utils.runtime(e);
		}
	}

	/**
	 * Set the working directory.
	 * 
	 * @param dir
	 *            Can be null for the JVM's working dir.
	 * @see ProcessBuilder#directory(File)
	 */
	public void setDirectory(File dir) {
		pb.directory(dir);
	}

	/**
	 * false by default. If true, also send all output to system out
	 * 
	 * @param echo
	 */
	public void setEcho(boolean echo) {
		assert process == null : "do earlier!";
		this.echo = echo;
	}

	/**
	 * If true, the error stream will be merged into the output stream.
	 * 
	 * @see ProcessBuilder#redirectErrorStream(boolean).
	 */
	public void setRedirectErrorStream(boolean redirect) {
		pb.redirectErrorStream(redirect);
	}

	@Override
	public String toString() {
		return pb.command().toString() + "\n" + out.getString();
	}

	/**
	 * Wait for the process to finish. If the thread is interrupted, this will
	 * destroy the process to ensure system resources are freed.
	 * 
	 * FIXME: encountered bug in {@link Utils#getSVNRevision(File)} where this
	 * appears to return too quickly
	 * 
	 * @return Exit code from the process, usually 0 for fine, other for error.
	 * @throws RuntimeException
	 *             if interrupted
	 */
	public int waitFor() {
		try {
			// attempting to patch a bug -- see ProcessTest.testBug()
			Utils.sleep(5);
			int v = process.waitFor();
			closeStreams();
			return v;
		} catch (InterruptedException e) {
			// Destroy the process!
			// Otherwise it might hog system resources, and the gobbler threads
			// will stay hung forever
			destroy();
			throw Utils.runtime(e);
		}
	}

	/**
	 * Wait for the process to finish. If the timeout is exceeded then the
	 * process will be destroyed to release any resources.
	 * 
	 * @param timeout
	 *            time in milliseconds to allow before giving up. 0 or -1
	 *            indicate to wait forever
	 * @return the exit code from the process. usually 0 for fine, other for
	 *         error.
	 * @throws wrapped
	 *             InterruptedException If the timeout is exceeded (or something
	 *             else interrupts).
	 */
	@SuppressWarnings("deprecation")
	public int waitFor(long timeout) {
		if (timeout < 1)
			return waitFor();
		try {
			// attempting to patch a bug -- see ProcessTest.testBug()
			Utils.sleep(5);
			TimeOut interrupter = new TimeOut(timeout);
			// Wait for it
			int v = process.waitFor();
			// Cancel that interruption
			interrupter.cancel();
			closeStreams();
			return v;
		} catch (InterruptedException e) {
			// Destroy the process!
			// Otherwise it might hog system resources, and the gobbler threads
			// will stay hung forever
			destroy();
			throw Utils.runtime(e);
		}
	}
}
