package winterwell.utils;

import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Level;

import winterwell.utils.reporting.Log;

/**
 * An interrupter to interrupt the thread that starts this. This typically leads
 * to an {@link InterruptedException} being thrown. Use to set a timeout on
 * blocking operations.
 * <p>
 * Typical usage:
 * 
 * <pre>
 * <code>
 * TimeOut timeOut = new TimeOut(millisecs);
 * try {
 * 	// do stuff
 * } finally {
 * 	timeOut.cancel();
 * }
 * </code>
 * </pre>
 * 
 * @author daniel
 */
public final class TimeOut {

	private static final Timer timer = new Timer("TimeOuter", true);
	private TimerTask task;
	private final long timeout;

	/**
	 * Create <i>and start</i> an interrupter to interrupt the thread that
	 * starts this. This typically leads to an {@link InterruptedException}
	 * being thrown. Use to set a timeout on blocking operations (which must be
	 * cancelled to avoid the interruption).
	 * 
	 * @param timeout
	 *            In milliseconds.
	 */
	public TimeOut(long timeout) {
		this.timeout = timeout;
		start();
	}

	/**
	 * @see TimerTask#cancel()
	 */
	public void cancel() {
		task.cancel();
	}

	private void start() {
		if (task != null)
			return; // already started
		final Thread target = Thread.currentThread();
		task = new TimerTask() {
			@Override
			public void run() {
				// log that we did indeed timeout
				Log.report("TimeOut on " + target.getName(), Level.WARNING);
				target.interrupt();
			}
		};
		timer.schedule(task, timeout);
	}

}
