package winterwell.utils.containers;

import java.io.Serializable;
import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;

import winterwell.utils.reporting.Log;

/**
 * A Set implementation backed by an {@link ArrayList}. Fast for small
 * collections, lousy for big ones. Has a convenient constructor. Another nice
 * property is that insertion order is preserved. Implements List and Set 'cos
 * it can.
 * 
 * FIXME {@link #equals(Object)} is List equals!
 * 
 * @author daniel
 * 
 * @param <T>
 */
public final class ArraySet<T> extends AbstractList<T> implements Set<T>,
		List<T>, Serializable {
	private static final long serialVersionUID = 1L;

	private final ArrayList<T> backing;

	public ArraySet() {
		backing = new ArrayList<T>();
	}

	public ArraySet(Collection<? extends T> elements) {
		backing = new ArrayList<T>(elements.size());
		for (T e : elements) {
			add(e);
		}
	}

	public ArraySet(int initialSize) {
		backing = new ArrayList<T>(initialSize);
	}

	public ArraySet(T... elements) {
		// This could be valid -- but it's probably a mistake
		if (elements.length == 1 && elements.getClass().isArray()) {
			Log.report("Probable bug: elements=[" + elements + "]",
					Level.WARNING);
		}
		backing = new ArrayList<T>(elements.length);
		for (T e : elements) {
			add(e);
		}
	}

	@Override
	public boolean add(T e) {
		if (backing.contains(e))
			return false;
		return backing.add(e);
	}

	@Override
	public void clear() {
		backing.clear();
	}

	@Override
	public boolean contains(Object o) {
		assert o == null || ((T) o) != null; // sanity check input class
		return backing.contains(o);
	}

	/**
	 * Since this is backed by an array-list, you might as well have indexed
	 * access.
	 * 
	 * @param i
	 * @return
	 */
	@Override
	public T get(int i) {
		return backing.get(i);
	}

	@Override
	public Iterator iterator() {
		return backing.iterator();
	}

	@Override
	public boolean remove(Object o) {
		assert o == null || ((T) o) != null; // sanity check input class
		return backing.remove(o);
	}

	/**
	 * This method does NOT implement set's no-duplicates check! Using this can
	 * allow duplicate entries.
	 */
	@Override
	@Deprecated
	public T set(int index, T element) {
		T old = backing.set(index, element);
		return old;
	}

	@Override
	public int size() {
		return backing.size();
	}

}
