package winterwell.utils.containers;

import java.util.HashMap;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

/**
 * Provides compare-and-set behaviour for a map, based on
 * {@link AtomicReference}.
 * 
 * 
 * @testedby {@link AtomicMapTest}
 * @author daniel
 */
public final class AtomicMap<K, V> extends AbstractMap2<K, V> {

	private final HashMap<K, AtomicReference<V>> backing = new HashMap<K, AtomicReference<V>>();

	private final Lock lock = new ReentrantLock();

	public boolean compareAndPut(K key, V expect, V update) {
		AtomicReference<V> ref = backing.get(key);
		if (ref == null) {
			// Is there a better way than having this one lock?
			// With one lock we'll get avoidable contention if the keys keep
			// changing
			lock.lock();
			ref = backing.get(key);
			if (ref == null) {
				ref = new AtomicReference<V>();
				backing.put(key, ref);
			}
			lock.unlock();
		}
		return ref.compareAndSet(expect, update);
	}

	@Override
	public V get(Object key) {
		AtomicReference<V> ref = backing.get(key);
		return ref == null ? null : ref.get();
	}

	@Override
	public Set<K> keySet() {
		return backing.keySet();
	}

	/**
	 * @deprecated This is a slow method. Use
	 *             {@link #compareAndPut(Object, Object, Object)} instead.
	 */
	@Deprecated
	@Override
	public V put(K key, V value) {
		synchronized (backing) {
			AtomicReference<V> ref = backing.get(key);
			V old;
			if (ref == null) {
				ref = new AtomicReference<V>();
				backing.put(key, ref);
				old = null;
			} else {
				old = ref.get();
			}
			ref.set(value);
			return old;
		}
	}

	@Override
	public V remove(Object key) {
		AtomicReference<V> ref = backing.remove(key);
		return ref.get();
	}

	@Override
	public int size() {
		return backing.size();
	}

}
