package winterwell.utils.io;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;
import java.text.DateFormat;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;

import winterwell.utils.ReflectionUtils;
import winterwell.utils.StrUtils;
import winterwell.utils.Utils;

/**
 * Parse Unix style command line arguments
 * 
 * Sets fields annotated with {@link Option} in a settings object.
 * 
 * @author daniel
 */
public class ArgsParser {

	static List<Class> recognisedTypes = Arrays.asList((Class) String.class,
			Double.class, Float.class, Integer.class, Long.class, double.class,
			float.class, int.class, long.class, Boolean.class, boolean.class,
			Date.class, File.class);

	/**
	 * Create an instance of type from a string representation.
	 * 
	 * @param type
	 * @param string
	 * @return
	 * @throws ParseException
	 */
	public static Object convert(Class<?> type, String string)
			throws ParseException {
		try {
			if (type == String.class)
				return string;
			// Numbers
			if (Number.class.isAssignableFrom(type)) {
				if (type == Double.class)
					return Double.valueOf(string);
				if (type == Float.class)
					return Float.valueOf(string);
				if (type == Integer.class)
					return Integer.valueOf(string);
				if (type == Long.class)
					return Long.valueOf(string);
			}
			if (type.isPrimitive()) {
				if (type == double.class)
					return Double.valueOf(string);
				if (type == float.class)
					return Float.valueOf(string);
				if (type == int.class)
					return Integer.valueOf(string);
				if (type == long.class)
					return Long.valueOf(string);
				if (type == boolean.class)
					return Boolean.valueOf(string);
			}
			// Date
			if (type == Date.class)
				return DateFormat.getInstance().parse(string);
			// File
			if (type == File.class)
				return new File(string);
			throw new IllegalArgumentException("Unrecognised type: " + type);
		} catch (Exception e) {
			throw new ParseException(e.getMessage(), 1);
		}
	}

	Map<Field, Object> field2default = new HashMap<Field, Object>();

	private List<Field> missingArgs;

	/**
	 * Reusable
	 */
	List<Field> requiredArgs = new ArrayList<Field>();

	/**
	 * Temporary copy used during a parse
	 */
	List<Field> requiredArgs2;

	private final Object settings;

	Map<String, Field> token2field = new HashMap<String, Field>();

	/**
	 * Create an ArgsParser which will set {@link Option} fields in the given
	 * settings object.
	 * 
	 * @param settings
	 */
	public ArgsParser(Object settings) {
		assert settings != null;
		this.settings = settings;
		// Setup token->field map
		parseSettingsObject(settings);
	}

	private void checkField(Class<?> type) {
		if (!recognisedTypes.contains(type))
			throw new IllegalArgumentException("Unrecognised type: " + type);
	}

	private void checkRequiredSettings(Object settings2) {
		if (requiredArgs2.size() == 0)
			return;
		missingArgs = requiredArgs2;
		throw new IllegalArgumentException();
	}

	public List<Field> getMissingArguments() {
		if (missingArgs == null)
			throw new IllegalStateException("parse has not run and failed");
		return missingArgs;
	}

	/**
	 * @return a message describing the available options.
	 */
	public String getOptionsMessage() {
		StringBuilder msg = new StringBuilder();
		msg.append("Options:" + StrUtils.LINEEND);
		HashSet<Field> options = new HashSet<Field>(token2field.values());
		for (Field f : options) {
			Option arg = f.getAnnotation(Option.class);
			String desc = arg.description();
			if (Utils.isBlank(desc)) {
				desc = f.getName();
			}
			msg.append("  " + tokens(f, arg) + '\t' + desc);
			if (field2default.get(f) != null) {
				msg.append(" Default: " + field2default.get(f));
			}
			msg.append(StrUtils.LINEEND);
		}
		return msg.toString();
	}

	/**
	 * @param args
	 *            The arguments as passed into main()
	 * @return the non-options arguments, i.e. the ones after the options
	 * @throws IOException
	 *             with a usage message. The missing arguments (if this is the
	 *             problem) can then be found via {@link #getMissingArguments()}
	 *             .
	 */
	public String[] parse(String[] args) throws IOException {
		requiredArgs2 = new ArrayList<Field>(requiredArgs);
		try {
			// Look for settings
			int i = 0;
			for (; i < args.length; i++) {
				String a = args[i];
				Field field = token2field.get(a);
				if (field == null) {
					break;
				}
				requiredArgs2.remove(field);
				if (field.getType() == Boolean.class
						|| field.getType() == boolean.class) {
					field.set(settings, true);
					// Did they specify?
					if (i + 1 < args.length) {
						String a2 = args[i + 1].toLowerCase();
						if (a2.equals("true") || a2.equals("false")) {
							field.set(settings, a2.equals("true"));
							i++;
						}
					}
				} else {
					// Take next argument as parameter
					i++;
					Object v = convert(field.getType(), args[i]);
					field.set(settings, v);
				}
			}
			// OK?
			checkRequiredSettings(settings);
			// return remainder
			return Arrays.copyOfRange(args, i, args.length);
		} catch (Exception e) {
			if (missingArgs == null) {
				missingArgs = new ArrayList<Field>(0);
			}
			throw new IOException(getOptionsMessage());
		}
	}

	/**
	 * setup the {@link #token2field} map
	 * 
	 * @param settings
	 * @throws IllegalArgumentException
	 */
	private void parseSettingsObject(Object settings)
			throws IllegalArgumentException {
		// Get annotated fields
		List<Field> fields = ReflectionUtils.getAnnotatedFields(settings,
				Option.class, true);
		// Get tokens
		for (Field field : fields) {
			// Is this OK?
			checkField(field.getType());
			// Get tokens
			Option arg = field.getAnnotation(Option.class);
			for (String t : tokens(field, arg).split(",")) {
				if (!t.startsWith("-"))
					throw new IllegalArgumentException(
							"Invalid token (all tokens must begin with a -): "
									+ t);
				token2field.put(t, field);
			}
			// Required?
			if (arg.required()) {
				requiredArgs.add(field);
			}
			// Default
			if (field.getType() != Boolean.class
					&& field.getType() != boolean.class) {
				try {
					if (!field.isAccessible()) {
						field.setAccessible(true);
					}
					Object d = field.get(settings);
					field2default.put(field, d);
				} catch (IllegalAccessException e) {
					throw new IllegalArgumentException(e);
				}
			}
			// ok
		}
	}

	private String tokens(Field f, Option arg) {
		String tokens = arg.tokens();
		if (tokens == null || tokens.length() == 0) {
			tokens = "-" + f.getName();
		}
		return tokens;
	}

	@Override
	public String toString() {
		return getClass().getSimpleName() + '\n' + getOptionsMessage();
	}

}
