package winterwell.utils.io;

import java.io.Closeable;
import java.io.File;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Iterator;

import winterwell.utils.StrUtils;
import winterwell.utils.Utils;
import winterwell.utils.reporting.Log;

/**
 * THIS CODE IS CURRENTLY INCOMPATIBLE WITH CSV WRITER: CsvWriter has been
 * modified to output "standard" CSV i.e. not using the escaping scheme
 * implemented below.
 * 
 * 
 * Support for reading comma (and tab) separated values.
 * <p>
 * By default, lines at the beginning of the file starting with a # are ignored
 * (assumed to be comments). Override {@link #isComment(String)} to change this.
 * 
 * Escaping is posix style - e.g. "hello, john" - rather than standard csv style
 * with double quotes ""hello, john"".
 * <p>
 * Will close when empty, or can be closed manually.
 * 
 * @testedby {@link CSVReaderTest}
 * @deprecated Currently incompatible with CSVWriter
 * @author daniel
 * 
 */
@Deprecated
public class OldCSVReader implements Iterable<String[]>, Iterator<String[]>,
		Closeable {

	/**
	 * constant for indicating rows can have a varying number of columns
	 */
	public static int VARIABLE_WIDTH = -2;
	int columns = -1;
	private final char delimiter;
	private final char escape = '\\';
	private boolean escaped;
	private String[] headers;
	private final LineReader input;
	private boolean inQuotes;
	private String[] next;
	private final char quote;

	private int rowNum;

	public OldCSVReader(File f, char delimiter) {
		this(f, delimiter, (char) 0);
	}

	public OldCSVReader(File file, char delimiter, char quote) {
		this(FileUtils.getReader(file), delimiter, quote);
	}

	/**
	 * Create a CSVReader with these settings. This can then be read using a
	 * foreach loop.
	 * 
	 * @param file
	 * @param delimiter
	 * @param quote
	 */
	public OldCSVReader(Reader input, char delimiter, char quote) {
		this.delimiter = delimiter;
		this.quote = quote;
		this.input = new LineReader(input);
		next = getNext();
	}

	@Override
	public void close() {
		input.close();
	}

	@Override
	protected void finalize() throws Throwable {
		close();
	}

	private String[] getFirstRow() {
		assert columns == -1;
		while (input.hasNext()) {
			ArrayList<String> row = new ArrayList<String>();
			String line = input.next();
			if (isComment(line)) {
				continue;
			}
			if (Utils.isBlank(line)) {
				continue;
			}
			getRow2_Fields(row, line);
			columns = row.size();
			return row.toArray(StrUtils.ARRAY);
		}
		return null;
	}

	/**
	 * @return the line number of the underlying LineReader.
	 * @see #getRowNumber()
	 */
	public int getLineNumber() {
		return input.getLineNum();
	}

	private String[] getNext() {
		return columns == -1 ? getFirstRow() : getRow();
	}

	private String[] getRow() {
		if (!input.hasNext())
			return null;
		assert columns > 0 || columns == VARIABLE_WIDTH;
		ArrayList<String> row = new ArrayList<String>(Math.max(columns, 1));
		boolean appending = false;
		int i = 0;
		while (input.hasNext()) {
			String line = input.next();
			// ?? support for blank lines inside quoted values
			if (Utils.isBlank(line)) {
				continue;
			}
			if (appending) {
				i = row.size();
			}
			getRow2_Fields(row, line);
			if (appending) {
				String app = row.remove(i);
				row.set(i - 1, row.get(i - 1) + StrUtils.LINEEND + app);
			}

			if (columns == VARIABLE_WIDTH) {
				break;
			}

			if (row.size() == columns) {
				break;
			}
			if (row.size() > columns) {
				reportBadRow(line);
				return getRow();
			}
			if (!inQuotes) {
				reportBadRow(line);
				return getRow();
			}
			appending = true;
		}
		if (columns != VARIABLE_WIDTH && row.size() != columns) {
			reportBadRow("eof!");
			return getRow();
		}
		rowNum++;
		return row.toArray(StrUtils.ARRAY);

	}

	private void getRow2_Fields(ArrayList<String> row, String line) {
		int col = 0;
		StringBuilder field = new StringBuilder();
		escaped = false;
		// ?? inQuotes=false?
		for (int i = 0, n = line.length(); i < n; i++) {
			char c = line.charAt(i);
			// add any char as-is if it is escaped
			if (escaped) {
				escaped = false;
				field.append(c);
				continue;
			}
			// escape the next char?
			if (c == escape) {
				assert !escaped;
				escaped = true;
				continue;
			}
			if (c == quote) {
				if (inQuotes) {
					// TODO escaped
					// A proper end-quotes should be at the end of the line or
					// then have a delimiter
					if (line.length() == i + 1
							|| line.charAt(i + 1) == delimiter) {
						inQuotes = false;
					} else {
						// a bad embedded quote
						field.append(c);
					}
				} else {
					// A proper start-quotes should be at the start of the line
					// or just after a delimiter
					if (i == 0 || line.charAt(i - 1) == delimiter) {
						inQuotes = true;
					} else {
						// a bad start-quote?
						field.append(c);
					}
				}
				continue;
			}
			if (c == delimiter && !inQuotes) {
				// Finished a field
				col++;
				inQuotes = false;
				escaped = false;
				row.add(field.toString());
				field = new StringBuilder();
				continue;
			}
			// just add it
			field.append(c);
		}
		// Add last field
		String f = field.toString();
		row.add(f);
	}

	/**
	 * The number of the last row returned. zero indexed (zero at the beginning
	 * too). This may be less than the line number due to comments and
	 * multi-line items
	 * 
	 * @see #getLineNumber()
	 */
	public int getRowNumber() {
		return rowNum;
	}

	@Override
	public boolean hasNext() {
		return next != null && next.length != 0;
	}

	protected boolean isComment(String line) {
		return line.startsWith("#");
	}

	@Override
	public Iterator<String[]> iterator() {
		return this;
	}

	@Override
	public String[] next() {
		String[] current = next;
		next = getNext();
		return current;
	}

	@Override
	public void remove() {
		throw new UnsupportedOperationException();
	}

	private void reportBadRow(String line) {
		// TODO something sensible
		Log.report("Bad row seen: " + line);
	}

	/**
	 * Fix the number of columns in this table (otherwise it will be set by the
	 * first row). Can be #VARIABLE_WIDTH
	 * 
	 * @param columns
	 */
	public void setNumberColumns(int columns) {
		assert columns > 0 || columns == VARIABLE_WIDTH;
		this.columns = columns;
	}

}
